MODULE Convert;
  (*
    Converter to Port Programs With Conditional Compilation
    Copyright (C) 1994, Alex Iakovlev

    NOTES:
    (1) Each conditional must be contained in SINGLE line.
    (2) TopSpeed's FALSE conditional aren't supported.
    (3) Stony Brook's %ELSIF, %ELSE, %NOT, %AND, %OR operators
        aren't supported.
  *)
  (*
    *************************************************************
    A.Iakovlev Ver  1.0  Initial version              February 94
    *************************************************************
  *)

  (*# debug(public => on) *)

  IMPORT ASCII, Lib, FIO, FIOR, IO, Str;

  CONST
    StrLen = 64;
    BuffLen = 1200;
    FileBuffLen = 24 * 512;
    Attr = FIO.FileAttr{};
    TempName = "CONVERT.$$$";

  TYPE
    String = ARRAY [0 .. StrLen - 1] OF CHAR;
    CondOps = RECORD begHead, begTail, end, tag: String END;
    DeltaOps = RECORD beg, end: INTEGER END;
    FileBuffer = ARRAY [0 .. FileBuffLen - 1] OF CHAR;
    Buffer = ARRAY [0 .. BuffLen - 1] OF CHAR;

  CONST
    TSCondOps = CondOps("(*%T ", "*)", "(*%E*)", "(*%");
    SBCondOps = CondOps("%IF ", " %THEN", "%END", "%");

  VAR
    dir, head, tail, filespec: String;
    dstBuff, srcBuff: FileBuffer;
    opsOld, opsNew: CondOps;
    delta: DeltaOps;
    ent: FIO.DirEntry;

  PROCEDURE ComputeDelta(VAR old, new: CondOps);
    BEGIN
      delta.beg :=
        INTEGER(Str.Length(new.begHead)) - INTEGER(Str.Length(old.begHead)) +
        INTEGER(Str.Length(new.begTail)) - INTEGER(Str.Length(old.begTail));
      delta.end :=
        INTEGER(Str.Length(new.end    )) - INTEGER(Str.Length(old.end    ));
    END ComputeDelta;

  PROCEDURE GetParameters(VAR fs : String; VAR old, new: CondOps): BOOLEAN;
    CONST
      TS2SB = "TS2SB";
      SB2TS = "SB2TS";
    VAR
      temp: String;
    BEGIN
      IF Lib.ParamCount() = 2 THEN
        Lib.ParamStr(fs, 1);
        Lib.ParamStr(temp, 2);
        Str.Caps(temp);
        IF Str.Compare(temp, TS2SB) = 0 THEN
          old := TSCondOps;
          new := SBCondOps;
          ComputeDelta(old, new);
          RETURN TRUE;
        ELSIF Str.Compare(temp, SB2TS) = 0 THEN
          old := SBCondOps;
          new := TSCondOps;
          ComputeDelta(old, new);
          RETURN TRUE;
        ELSE
          RETURN FALSE;
        END;
      ELSE
        RETURN FALSE;
      END;
    END GetParameters;

  PROCEDURE SubstTail(VAR b: Buffer; pos: CARDINAL; old, new: String);
    VAR
      pos1: CARDINAL;
      head: String;
    BEGIN
      pos1 := Str.Pos(b, old);
      IF pos1 >= pos THEN
        (* ------------- tail is found AFTER pos *)
        Str.Subst(b, old, new);
      ELSE
        (* ----------------- non-correct tail: split the line *)
        Str.Slice(head, b, 0, pos - 1);
        Str.Delete(b, 0, pos - 1);
        (* ------------------ substitute tail of beginning conditional *)
        Str.Subst(b, old, new);
        (* ----------------------------- combine the line again *)
        Str.Prepend(b, head);
      END;
    END SubstTail;

  PROCEDURE SubstLine(VAR b: Buffer; VAR len: CARDINAL; VAR old, new: CondOps): CARDINAL;
    VAR
      count, pos: CARDINAL;
    BEGIN
      IF len > 0 THEN
        (* ------------------ line isn't zero-length *)
        count := 0;
        LOOP
          pos := Str.Pos(b, old.begHead);
          IF pos < MAX(CARDINAL) THEN
            (* ---------------- beginning conditional is found *)
            Str.Subst(b, old.begHead, new.begHead);
            SubstTail(b, pos, old.begTail, new.begTail);
            len := INTEGER(len) + delta.beg;
            INC(count);
          ELSIF Str.Pos(b, old.end) < MAX(CARDINAL) THEN
            (* ---------------- ending conditional is found *)
            Str.Subst(b, old.end, new.end);
            len := INTEGER(len) + delta.end;
            INC(count);
          ELSE
            (* -------------- no conditionals *)
            EXIT;
          END;
        END;
        RETURN count;
      ELSE
        (* ------- line is too short *)
        RETURN 0;
      END;
    END SubstLine;

  PROCEDURE RdLine(f: FIO.File; VAR b: Buffer; VAR len: CARDINAL);
    VAR
      ch: CHAR;
    BEGIN
      FIO.RdStr(f, b);
      len := Str.Length(b);
      (* ----------------------- EOL/EOF processing *)
      IF FIO.EOF THEN
        b[len] := CHR(26);
        INC(len);
      ELSE
        b[len] := CHR(13);
        INC(len);
        b[len] := CHR(10);
        INC(len);
      END;
      (* ----------------- NULL *)
      b[len] := CHR(0);
    END RdLine;

  PROCEDURE WrLine(f: FIO.File; b: Buffer; len: CARDINAL);
    BEGIN
      FIO.WrBin(f, b, len);
    END WrLine;

  PROCEDURE SubstFile(dstName, srcName: String; VAR old, new: CondOps): CARDINAL;
    VAR
      buff: Buffer;
      count, len, oldTagLen: CARDINAL;
      src, dst: FIO.File;
    BEGIN
      FIO.IOcheck := FALSE;
      dst := FIO.Create(dstName);
      FIO.AssignBuffer(dst, dstBuff);
      src := FIO.OpenRead(srcName);
      FIO.AssignBuffer(src, srcBuff);
      oldTagLen := Str.Length(old.tag);
      count := 0;
      REPEAT
        RdLine(src, buff, len);
        IF (len >= oldTagLen) AND (Str.Pos(buff, old.tag) < MAX(CARDINAL)) THEN
          INC(count, SubstLine(buff, len, old, new));
        END;
        WrLine(dst, buff, len);
      UNTIL FIO.EOF;
      FIO.Close(src);
      FIO.Erase(srcName);
      FIO.Truncate(dst);
      FIO.Close(dst);
      FIO.IOcheck := TRUE;
      RETURN count;
    END SubstFile;

  PROCEDURE CopyAndFound(dstName, srcName, oldTag: String): BOOLEAN;
    VAR
      buff: Buffer;
      src, dst: FIO.File;
      len, oldTagLen: CARDINAL;
      found: BOOLEAN;
    BEGIN
      FIO.IOcheck := FALSE;
      dst := FIO.Create(dstName);
      FIO.AssignBuffer(dst, dstBuff);
      src := FIO.OpenRead(srcName);
      FIO.AssignBuffer(src, srcBuff);
      oldTagLen := Str.Length(oldTag);
      found := FALSE;
      REPEAT
        RdLine(src, buff, len);
        IF NOT found AND (len >= oldTagLen) THEN
          found := (Str.Pos(buff, oldTag) < MAX(CARDINAL));
        END;
        WrLine(dst, buff, len);
      UNTIL FIO.EOF;
      FIO.Close(src);
      FIO.Close(dst);
      FIO.IOcheck := TRUE;
      RETURN found;
    END CopyAndFound;

  PROCEDURE TextFile(name: ARRAY OF CHAR): BOOLEAN;
    CONST CharNmb = 100;
    VAR
      src: FIO.File;
      buff: ARRAY [0 .. CharNmb - 1] OF CHAR;
      read, j: CARDINAL;
      ok: BOOLEAN;
      ch: CHAR;
    BEGIN
      FIO.IOcheck := FALSE;
      src := FIO.OpenRead(name);
      FIO.AssignBuffer(src, srcBuff);
      read := FIO.RdBin(src, buff, CharNmb);
      IF read > 0 THEN
        j := 0;
        ok := TRUE;
        WHILE (j < read) AND ok DO
          ch := buff[j];
          ok := (ch >= " ") OR (ch = ASCII.cr) OR (ch = ASCII.lf) OR (ch = ASCII.sub);
          INC(j);
        END;
      ELSE
        ok := FALSE;
      END;
      FIO.Close(src);
      FIO.IOcheck := TRUE;
      RETURN ok;
    END TextFile;

  BEGIN
    IF GetParameters(dir, opsOld, opsNew) THEN
      IF FIO.ReadFirstEntry(dir, Attr, ent) THEN
        FIOR.SplitPath(dir, head, tail);
        REPEAT
          FIOR.MakePath(filespec, head, ent.Name);
          IF TextFile(filespec) AND CopyAndFound(TempName, filespec, opsOld.tag) THEN
            IO.WrStr("   File : ");
            IO.WrStrAdj(ent.Name, SIZE(FIO.PathTail) + 2);
            IO.WrCard(SubstFile(filespec, TempName, opsOld, opsNew), 8);
            IO.WrStr(" substitutions");
            IO.WrLn();
          ELSE
            FIO.Erase(TempName);
          END;
        UNTIL NOT FIO.ReadNextEntry(ent);
      ELSE
        IO.WrStr("Directory ");
        IO.WrStr(filespec);
        IO.WrStr(" doesn't contain any files");
        IO.WrLn();
      END;
    ELSE
      IO.WrStr("Usage: CONVERT <filespec> ( ts2sb | sb2ts )");
      IO.WrLn();
    END;
  END Convert.
